#! /usr/bin/env python3

import sys
sys.path.insert(0,"/opt/ark-ping-watch/lib")
import plt
import os
import re
import stat
import sqlite3
import pickle
import libtiny

from time				import time, strftime
from json				import dumps
#from Crypto.PublicKey	import RSA
from ark_ping_sqlite	import StateDB, HistoryDB
from ark_ping_layout	import validate_attributes

private_key = None

#+
# NAME:
#	ark-ping-watch.py
# PURPOSE:
#	Captures 'ping-home' packets sent hourly to indy by Ark nodes.
# CALLING SEQUENCE:
#	ark-ping-watch.py --ip [ip] --interface [network interface] --db [db]
# OPTIONAL INPUTS:
#	-n,--dry-run
#	    suppresses sending of email (so is the same as just
#	    not using --email)
#	-v,--verbose
#	--debug <n>
#	    increase verbosity (--verbose is same as --debug 1)
# RESTRICTIONS:
#	wandio-4.2.1.tar.gz
#	libtrace-4.0.7.tar.gz
#	pypy-libtrace-master.zip
#
#	https://docs.python.org/3/library/queue.html
#	https://search.auckland.ac.nz/search?site=All_university_websites&q=libtrace
#	https://github.com/nevil-brownlee/python-libtrace
#
#	Needs sudo access.
#
#	On indy a previous verison of this script is run in a tmux session in account ark:
#		export PYTHONPATH=$HOME/toolshed/lib
#		cd /data/topology/ark/ping-home
#		$HOME/toolshed/bin/ark-ping-watch --log ./ping-home.log --db ./ping-state.db --verbose
#	On sirocco, this script is run in a tmux session in account ark:
#		export PYTHONPATH=$HOME/toolshed/lib
#		cd /data/topology/ark/ping-home
#		$HOME/toolshed/bin/ark-ping-watch --ip 192.172.226.86 --interface ens6 --db ./ping-state.db --debug 2
#	ark-ping-watch has setuid root. Note:
#	- to capture packets root access is needed on indy (hence the wrapper)
#	- the log file is owned by ark (so use 'sudo -u ark tee' if you are not
#	  on the ark account.
# PROCEDURE:
#	Packets are captured using the Python plt module, or, if that is not
#	available, by directly calling tcpdump.
#
#	In tcpdump packets look like:
#	16:32:01.991700 IP 173.76.237.173 > 192.172.226.37: ICMP echo request, id 6437, seq 1, length 26
#		0x0000: 4500 002e 57d1 4000 2d01 b831 ad4c edad E...W.@.-..1.L..
#		0x0010: c0ac e225 0800 077c 1925 0001 d9e4 5e53 ...%...|.%....^S
#		0x0020: 09c7 0500 2928 6265 6432 2d75 7329 ....)(bed2-us)
#	i.e. the source IP address is in the first line; the hostname in
#	the last. These are extracted using regular expressions.
#
#	Summaries of packets received are either send as emails and/or printed
#	to stderr.
#
#	Packets are received in a roughly two minute window between hh:30 and
#	hh:32 every hour. The html summary always happens when the first packet
#	is received. For san-us the packet is intentionally delayed by 3
#	minutes, so its the last packet to arrive each hour.
#
#	If a hostname changes on existing hardware then ideally the history under the
#	the new hostname is appended to the history under the old name, i.e. the
#	hostname in the state db is changed to the new value, and the history file
#	is renamed. There are three ways to accomplish this.
#
#	1. The first packet send from the node after the name change contains the
#	   mac address of the network interface. An attempt is made to find the
#	   node in the state db with the same mac address.
#	2. The first packet send from the node after the name change is a
#	   custom 'type 99' packet which explicitly specifies the old name:
#
#	   sudo python3 /usr/local/ark/bin/ark-ping-beacon.py --jitter 0 [--node NEW_NODE] Hostname=OLD_NODE
#
#	   --node NEW_NODE is optional. If not specified the ARK_MONITOR value in
#	   /usr/local/ark/config/monitor-config is used, or else 'hostname -s'.
#
#	3. Before the node sends its first ping-home packet (which may mean
#	   before booting the node under the new name), send a custom 'type 99'
#	   packet from any machine inside the CAIDA/24:
#
#	   sudo python3 /usr/local/ark/bin/ark-ping-beacon.py --jitter 0 --node NEW_NODE Hostname=OLD_NODE
#
#	   This time --node NEW_NODE is mandatory.
#
#	If neither of these three is done in time, then a new node is added
#	to the state db with a new history file. The records for the old node name will
#	eventually be removed (after six months of not seeing any packets), and the
#	history file is preserved.
#
# MODIFICATION HISTORY:
#	MAY-2014, Paul Hick (UCSD/CAIDA)
#	SEP-2014, Paul Hick (UCSD/CAIDA)
#		Added processing of .monitors.sh to process 'inactive monitors' (listed
#		in .monitors.sh, but not in monitors.yaml
#	MAY-2015, Paul Hick (UCSD/CAIDA)
#		Added --alert-on-ping
#		Removed use of obsolete .monitors.sh. Instead get inactiver monitors
#	`	from monitors-bonus.yaml
#	SEP-2015, Paul Hick (UCSD/CAIDA; pphick@caida.org), V1.02
#		Added check for public IPv6 address on interface eth0 (i.e. works
#		only for Raspberry Pis).
#	JUN-2019, Paul Hick (UCSD/CAIDA; pphick@caida.org), V1.10
#		Removed a bunch of tasks: html summary, alert on ping
#	MAR-2023, Victoria Nguyen (UCSD/CAIDA; vin011@ciada.org) 
#		Replaced orgparse with argparse. Converted network interface and hostname 
#		address values to be passed via required arguments
#-

def start_capture( interface, tcpdump_filter ):
	say = libtiny.say('start_capture')
	say.debug( interface+": "+tcpdump_filter )
	trace = plt.trace("int:" + interface)
	filter = plt.filter( tcpdump_filter )
	trace.conf_filter(filter)
	trace.conf_snaplen(65536)
	trace.conf_promisc(False)
	trace.start()
	return trace

def pkt_reader( interface, dest_address ):
	say = libtiny.say('pkt_reader')

	# yields pkt dict:
	#
	#	boot packet:
	#
	#	'time'
	#	'node'
	#	'IPv4Global'
	#	'IPv6Global'
	#	'Teredo'
	#
	#	network packet
	#
	#	'time'
	#	'node'
	#	'TX'
	#	'RX'
	#
	#	activity packet
	#
	#	'time'
	#	'node'
	#	'activity'
	#
	#	default packet:
	#
	#	'time'
	#	'node'
	#	'IPv4Global'
	#	'IPv6Global'
	#	'Teredo'
	#	'uptime in secs'
	#	'utils (0|1|2)*7'
	#

	#           node name
	re_mon = '([a-z][a-z,0-9]{2}\d?(?:-[a-z]{2})?)'

	re_base = re.compile('\(%s\|\d\|'%re_mon)

	#                          (Mon|0 | IPv4 |Teredo       |IPv6         |Uptime| Utils )
	re_default  = re.compile('\(%s\|0\|(.*?)\|(.*?)\|(.*?)\|(.*?)\|(.*?)\|(.*?)\|(\d+?)\)'%re_mon)

	#                          (Mon|1 | IPv4 |Teredo       |IPv6         |OSname   |Mac    )
	re_boot     = re.compile('\(%s\|1\|(.*?)\|(.*?)\|(.*?)\|(.*?)\|(.*?)\|(.*?)(?:\|(.*))?\)'%re_mon)
	#(xxx-yy|1|192.172.226.249|2001:0:c0ac:e290:5c:47cc:3f53:1d06|2001:48d0:101:501:ba27:ebff:fea7:dd86)

	#                          (Mon|2 |  TX   |  RX   |Rate )
	re_network  = re.compile('\(%s\|2\|(\d+?)\|(\d+?)\|(\d{4}/\d{2}/\d{2}=.+)\)'	%re_mon)
	#                          (Mon|3 |  Avail  /  Used  / Size  |Dirs |List)
	re_activity = re.compile('\(%s\|3\|(?:(\d+?)/)?(\d+?)/(\d+?)\|(.*)\|(.*)\)'		%re_mon)
	#                          (Mon|4 |Offset   |Reach |Server                                  |Server)
	re_ntp      = re.compile('\(%s\|4\|-?(\d+?)\|(\d+)\|(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})?(?:\|(.*))?\)'%re_mon)
	#                          (Mon|5 |IsUp   |Server)
	re_dns      = re.compile('\(%s\|5\|(\d+?)\|(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})\)'%re_mon)
	# (cld3-us|6|Raspbian GNU/Linux 10 (buster)|Linux 4.19.42-v7+|Raspberry Pi 3 Model B Rev 1.2|b8:27:eb:4c:27:63|999036|BCM2835|a22082|00000000434c2763|armv7l|192.168.0.1|c4:04:15:54:fa:0e|0x000003|0x1a2a7f20|15120888)
	# (bcn-es|6|FreeBSD 10.1|FreeBSD 10.1-RELEASE-p35|||2117103|||||84.88.81.121|00:12:1e:23:70:db|||471012072)
	#                              (Mon|6 |OSname|OSkrnl|Model|MAC |Mem  |Chip |Ver |Serial|Arm |GWIP  |GWMac|SDMID|SDSN |SDcapacity)
	re_hardware	= re.compile('\(%s\|6\|(.+?)\|(.+?)\|(.*?)\|(.*?)\|(.*?)\|(.*?)\|(.*?)\|(.*?)\|(.*?)\|(.+?)\|(.+?)(?:\|(.*)\|(.*)\|(.*))?\)'%re_mon)

	re_storage	= re.compile('\(%s\|7\|(.+?)\|(.+?)\|(.*?)\|(.*?)\|(.*?)\|(.*?)\|(.*?)\)'%re_mon)
	#                                                                       key    value
	re_send		= re.compile('\(%s\|99\|(.*)\)'%re_mon)

	tcpdump_filter = ('icmp' if re.match('^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$',dest_address) else 'icmp6') + ' and dst '+dest_address

	last_posix_time = None
	last_payload	= None

	for pkt in start_capture( interface, tcpdump_filter ):

		# Only pick up IPv4 ICMP type 8, and IPv6 ICMP type 128

		if not ((pkt.ip and pkt.icmp.type == 8) or (pkt.ip6 and pkt.icmp6.type == 128)):
			continue

		# The following try-except is necessary to work around issues with the py-libtrace library and also 
		# covers cases of malformed packets
		try:
			# The first line works on indy; the second for newer versions of plt.
			#payload = str( pkt.icmp.payload.data if pkt.ip else pkt.icmp6.payload.data )
			payload = str( pkt.icmp.payload if pkt.ip else pkt.icmp6.payload )
		except ValueError:
			log.write(f"{strftime('%Y-%m-%dT%H:%M:%S')}: skipped packet: prefix {pkt.ip.src_prefix}, {pkt.layer3.data.hex()}\n")
			continue

		if private_key:
			payload = private_key.decrypt(payload)

		posix_time = round(time())

		if last_posix_time and posix_time == last_posix_time and last_payload == payload:
			say.deep_debug( 'reject duplicate packet' )
			continue

		last_posix_time = posix_time
		last_payload	= payload

		payload = payload[12:-2]			# "bytearray(b'(...)')"
		message = '%s %15s %s'%(posix_time, str(pkt.ip.src_prefix) if pkt.ip else 'None', payload)

		m = re_boot.search( payload )
		if m:
			yield {
				'message'		: message		,
				'time'			: posix_time	,
				'node'			: m.group(1)	,
				'IPv4Global'	: str(pkt.ip.src_prefix) if pkt.ip else 'None',

				'PktType'		: 1				,
				'IPv4Local'		: m.group(2) if m.group(2) else 'None'	,
				#'Teredo'		: m.group(3) if m.group(3) else 'None'	,
				#'PingTeredo'	: m.group(4) if m.group(4) else 'None'	,
				'IPv6Global'	: m.group(5) if m.group(5) else 'None'	,
				'PingIPv6'		: m.group(6) if m.group(6) else 'None'	,
				'OSkernel'		: m.group(7) if m.group(7) else 'None'	,
				'HWmac'			: m.group(8) if m.group(8) else 'None'	,
			}
			continue

		m = re_network.search( payload )
		if m:
			yield {
				'message'		: message		,
				'time'			: posix_time	,
				'node'			: m.group(1)	,
				'IPv4Global'	: str(pkt.ip.src_prefix) if pkt.ip  else 'None',

				'PktType'		: 2				,
				'TX'			: m.group(2)	,		# Always contains a number
				'RX'			: m.group(3)	,		# Always contains a number
				'DailyRate'		: m.group(4)	,		# Empty on FreeBSD
			}
			continue

		m = re_activity.search( payload )
		if m:
			yield {
				'message'		: message		,
				'time'			: posix_time	,
				'node'			: m.group(1)	,
				'IPv4Global'	: str(pkt.ip.src_prefix) if pkt.ip  else 'None',

				'PktType'		: 3				,
				'SDAvail'		: m.group(2) if m.group(2) else '0'		,
				'SDUsed'		: m.group(3)	,		# Always contains a number
				'SDSize'		: m.group(4)	,		# Always contains a number
				'ActivityDirs'	: m.group(5) if m.group(4) else 'None'	,
				'ActivityList'	: m.group(6) if m.group(5) else 'None'	,
			}
			continue

		m = re_ntp.search( payload )
		if m:
			yield {
				'message'		: message		,
				'time'			: posix_time	,
				'node'			: m.group(1)	,
				'IPv4Global'	: str(pkt.ip.src_prefix) if pkt.ip  else 'None',

				'PktType'		: 4				,
				'NtpOffset'		: m.group(2)	,		# Always contains a number
				'NtpWorks'		: m.group(3)	,		# Always contains a number (0-377)
				'NtpServer'		: m.group(4) if m.group(5) else 'None'	,
				'NtpServerCf'	: m.group(5) if m.group(5) else 'None'	,
			}
			continue

		m = re_dns.search( payload )
		if m:
			yield {
				'message'		: message		,
				'time'			: posix_time	,
				'node'			: m.group(1)	,
				'IPv4Global'	: str(pkt.ip.src_prefix ) if pkt.ip  else 'None',

				'PktType'		: 5				,
				'DnsWorks'		: m.group(2)	,		# Always contains a number
				'DnsServer'		: m.group(3) if m.group(3) else 'None'	,
			}
			continue

		m = re_hardware.search( payload )
		if m:
			yield {
				'message'		: message		,
				'time'			: posix_time	,
				'node'			: m.group(1)	,
				'IPv4Global'	: str(pkt.ip.src_prefix) if pkt.ip  else 'None',

				'PktType'		: 6				,
				'OSname'		: m.group( 2) if m.group( 2) else 'None'	,
				'OSkernel'		: m.group( 3) if m.group( 3) else 'None'	,
				'HWmodel'		: m.group( 4) if m.group( 4) else 'None'	,
				'HWmac'			: m.group( 5) if m.group( 5) else 'None'	,
				'HWmem'			: m.group( 6) if m.group( 6) else 'None'	,
				'CPUchip'		: m.group( 7) if m.group( 7) else 'None'	,
				'CPUrevision'	: m.group( 8) if m.group( 8) else 'None'	,
				'CPUserial'		: m.group( 9) if m.group( 9) else 'None'	,
				'CPUarch'		: m.group(10) if m.group(10) else 'None'	,
				'GWIPv4'		: m.group(11) if m.group(11) else 'None'	,
				'GWmac'			: m.group(12) if m.group(12) else 'None'	,
				#'SDManfID'		: m.group(13) if m.group(13) else 'None'	,
				#'SDSerial'		: m.group(14) if m.group(14) else 'None'	,
				#'SDCapacity'	: m.group(15) if m.group(15) else 'None'	,
			}
			continue

		m = re_storage.search( payload )
		if m:
			yield {
				'message'		: message		,
				'time'			: posix_time	,
				'node'			: m.group(1)	,
				'IPv4Global'	: str(pkt.ip.src_prefix) if pkt.ip  else 'None',

				'PktType'		: 7				,
				'SDManfID'		: m.group( 2) if m.group( 2) else 'None'	,
				'SDModel'		: m.group( 3) if m.group( 3) else 'None'	,
				'SDSerial'		: m.group( 4) if m.group( 4) else 'None'	,
				'SDDate'		: m.group( 5) if m.group( 5) else 'None'	,
				'SDCapacity'	: m.group( 6) if m.group( 6) else 'None'	,
				'SDBootPart'	: m.group( 7) if m.group( 7) else 'None'	,
				'SDRootPart'	: m.group( 8) if m.group( 8) else 'None'	,
			}
			continue

		m = re_send.search( payload )
		if m:
			boat = {
				'message'		: message		,
				'time'			: posix_time	,
				'node'			: m.group(1)	,
				'sender'		: str(pkt.ip.src_prefix) if pkt.ip  else 'None',	# NOT IPv4Global!!

				'PktType'		: 99			,
			}

			attr = []
			for kv in m.group(2).split('|'):
				p = kv.find('=')
				k = kv[0:p]
				v = kv[p+1:]
				if k == 'time':
					v = int(v)
				else:
					attr.append(k)
				boat[k] = v if v else 'None'

			if validate_attributes( attr, exit=False ):
				yield boat
			continue

		m = re_default.search( payload )
		if m:
			yield {
				'message'		: message		,
				'time'			: posix_time	,
				'node'			: m.group(1)	,
				'IPv4Global'	: str(pkt.ip.src_prefix ) if pkt.ip else 'None'									,

				'PktType'		: 0				,
				'IPv4Local'		: m.group(2) if pkt.ip and m.group(2) else 'None'								,
				#'Teredo'		: str(pkt.ip6.src_prefix) if pkt.ip6 else m.group(3) if m.group(3) else 'None'	,
				#'PingTeredo'	: m.group(4) if pkt.ip and m.group(4) else 'None'								,
				'IPv6Global'	: m.group(5) if pkt.ip and m.group(5) else 'None'								,
				'PingIPv6'		: m.group(6) if pkt.ip and m.group(6) else 'None'								,
				'Uptime'		: m.group(7)	,	# Always contains a number
				'Utils'			: m.group(8)	,	# Always contains a number
			}
			continue

		say.deep_debug( 'drop packet %s'%message )

def check_pkt( pkt ):
	say = libtiny.say('check_pkt')

	# Check for NAT by comparing native and public IPv4 addresses

	if 'IPv4Global' in pkt:
		if 'IPv4Local' in pkt:
			pkt['NAT'] = '0' if pkt[ 'IPv4Global' ] == pkt[ 'IPv4Local' ] else '1'

	if 'IPv6Global' in pkt and pkt['IPv6Global'] != 'None':
		if 'PingIPv6' in pkt:
			pkt['PingIPv6'  ] = '1' if pkt['PingIPv6'  ] == '0' else '0'
	elif 'PingIPv6' in pkt:
		if pkt['IPv6Global'] == 'None':
			pkt['PingIPv6'] = 'None'
		else:
			del pkt['PingIPv6'  ]
		
	#if 'Teredo' in pkt and pkt['Teredo'] != 'None':
	#	if 'PingTeredo' in pkt:
	#		pkt['PingTeredo'] = '1' if pkt['PingTeredo'] == '0' else '0'
	#elif 'PingTeredo' in pkt:
	#	if pkt['Teredo'] == 'None':
	#		pkt['PingTeredo'] = 'None'
	#	else:
	#		del pkt['PingTeredo']

	pkt['Boot'] = '1' if pkt['PktType'] == 1 else '0'

	# Every packet will have pkt['node'] (from the payload)
	# pkt['Hostname'] only shows up in type 99 to mark 
	# a hostname change.

	if 'Hostname' not in pkt:
		pkt['Hostname'] = pkt['node']

	say.deep_debug( '%s'%pkt )

	return

def handle_changes( task_path, pkt, changes ):
	say = libtiny.say('handle_changes')

	# PktType=1 is boot packet
	if changes and (pkt['PktType'] == 1 or len( set(state_db.tbl_task).intersection(set(changes.keys())) ) > 0):
		file = os.path.join( task_path, str(int(round(time())))+'.'+pkt['node']+'.task' )
		open(file,'w').write( dumps( changes )+'\n' )

def load_history( db, node ):
	say = libtiny.say('load_history')

	# Every node has an IPv4Global address
	# If there is none in the state db it is a new node or
	# a node that was tossed out of the state db.
	# Retrieve history from the history db if it exists

	if not db.node_path:
		return

	node_db_file = os.path.join( db.node_path, node+'.db' )
	if not os.path.exists( node_db_file ):
		return					# No history: new node

	node_db = HistoryDB( node_db_file, db.tbl_names )
	if not node_db:
		return

	n = 0
	for tbl in db.tbl_names:
		tt, vv = db.find_value_for_node( tbl, node, node_db )
		if tt != None:
			n += 1
			if vv:
				db.insert( tbl, node, tt, vv )

	if n > 0:
		say.say( 'for node %s'%node )

if __name__ == '__main__':

	from argparse import ArgumentParser

	usage = "prog [-h] [arguments]"

	parser = ArgumentParser(usage=usage)

	parser.add_argument('--ip'			,
		dest		= 'hostname_ip'		,
		action		= 'store'			,
		help		= 'hostname ip address specification'	,
		required	= True				,
	)

	parser.add_argument('--interface'	,
		dest		= 'interface'		,
		action		= 'store'			,
		help		= 'network interface specification'	,
		required	= True				,
	)

	parser.add_argument('--db'			,
		dest		= 'db'				,
		action		= 'store'			,
		#default	= '/data/topology/ark/ping-home/ping-state.db'	,
		help		= 'sqlite db with current node attributes'		,
		required	= False				,
	)

	parser.add_argument('-v', '--verbose'	,
		dest		= 'verbose'			,
		action		= 'store_true'		,
		default		= False				,
		help		= 'verbose output'	,
		required	= False				,
	)

	parser.add_argument('--debug'		,
		dest		= 'debug'			,
		action		= 'store'			,
		type		= int				,
		default		= 0					,
		help		= 'set debug level'	,
		required	= False				,
	)

	parser.add_argument('-n', '--dry-run'	,
		dest		= 'dryrun'			,
		action		= 'store_true'		,
		default		= False				,
		help		= 'make dryrun'		,
		required	= False				,
	)

	parser.add_argument('-t', '--trigger'	,
		dest		= 'trigger'			,
		action		= 'store'			,
		help		= 'nr of secs used to trigger writes to db'	,
		required	= False				,
	)

	parser.add_argument('-c', '--count'	,
		dest		= 'count'			,
		action		= 'store'			,
		type		= int				,
		default		= 0					,
		help		= 'max number of packets to capture (default: no limit)'	,
		required	= False				,
	)

	parser.add_argument('-l', '--log'	,
		dest		= 'log'				,
		action		= 'store'			,
		default		= ''				,
		help		= 'log file for incoming packets'	,
		required	= False				,
	)

	parser.add_argument('--decrypt'		,
		dest		= 'decrypt'			,
		action		= 'store'			,
		default		= ''				,
		help		= 'private key file',
		required	= False				,
	)

	args = parser.parse_args()

	if args.count < 0:
		parser.error("--count (number of packets to capture) must be positive")

	say = libtiny.say(
		label	= 'ark-ping-watch'				,
		verbose = max(args.verbose,args.debug)	,
		dryrun  = args.dryrun					,
	)

	if args.decrypt:
		from Cryptodome.PublicKey import RSA
		private_key_file = open( args.decrypt, 'r')
		private_key = pickle.load(private_key_file)
		private_key_file.close()

	log = sys.stdout if args.log == '' else open( args.log, 'a+' )

	trigger_time  = time()
	trigger_delta = None

	state_db = StateDB( args.db, create_tables=True )
	have_state = True if state_db.db else False

	if state_db.db:
		if state_db.node_path:
			say.debug('history directory, %s/'%state_db.node_path)

		if state_db.task_path:
			say.debug('task directory, %s'%state_db.task_path)

		if not args.trigger:
			trigger_delta = 86400
			say.debug('trigger delta is %s s'%trigger_delta)
		elif args.trigger.isdigit():
			trigger_delta = int(args.trigger)
			say.debug('trigger delta is %s s'%trigger_delta)

	# On indy the main interface is em1 with IPv4 address '192.172.226.134'
	# On malibu the interface is eth0
	# On sirocco the interface is ens6 with address '192.172.226.86'

	#interface = args[0] if len(args) > 0 else 'em1'
	interface = args.interface
	if len( (os.popen('ifconfig %s 2> /dev/null'%interface)).read() ) == 0:
		say.die("interface '%s' does not exist?"%interface)
	#dest_address = os.popen('hostname --all-ip-addresses').read()[0:-1].split()[0]
	dest_address = args.hostname_ip
	say.debug("listen for dest '%s' on interface '%s'"%(dest_address,interface))

	pkt_count = 0
	mon_count = 0

	previous_nodes = set()
	current_nodes  = set()

	try:
		for pkt in pkt_reader( interface, dest_address ):

			check_pkt( pkt )
			log.write( pkt['message']+"\n" )
			log.flush()

			posix_time	= pkt[ 'time' ]
			node		= pkt[ 'node' ]

			# Each packet contains pkt['node'] and pkt['Hostname'].
			# pkt['node'] is extracted from the payload.  For all packets coming
			# from Ark nodes pktpkt['Hostname'] = pkt['node'].

			if pkt[ 'PktType' ] == 99:			# Accept custom db addition?

				# Only for pkt['PktType']=99 can 'Hostname' be different from pkt['node']
				# to indicate a hostname change. In that case pkt['Hostname'] is the
				# old (existing) node name that needs to be renamed to pkt['node'].

				if have_state:

					# If pkt['Hostname'] is not in the state db, then it is either a new
					# node (no history db), or a node that was inactive for so long it was
					# removed. In the second case a history db should exist, so repopulate
					# the state db with the most recent data from the history db.

					if not state_db.node_has_state( pkt['Hostname'] ):
						load_history ( state_db, pkt['Hostname'] )

					# If pkt['Hostname'] is still not in the state db, then drop the packet
					# (do NOT allow type 99 packets to create a new node entry).

					if not state_db.node_has_state( pkt['Hostname']):
						continue

					# At this point pkt['Hostname'] is in the state db.

					# Custom (type 99) packets are accepted only if they come from the
					# CAIDA 192.172.226.0/24, or if they come from an IP address already
					# in the state db (i.e. from an Ark node)

					if not pkt['sender'].startswith('192.172.226.'):	# Not from CAIDA/24?
						tt, vv = state_db.find_node_for_value( 'IPv4Global', pkt['sender'] )
						if tt == None:					# In state db?
							say.say ( 'reject (no IPv4Global on file): ', pkt )
							continue

			else:		# These packets are coming from an Ark node

				# Drop packets from CAIDA /24 (probably pi2.ark or some test device).
				# Keep this here (after check_pkt); we do want to see the packets in the
				# log file, just not added to the db.

				if pkt['IPv4Global'].startswith('192.172.226.'):		# From CAIDA /24?

					# san-us is inside the CAIDA/24; xxx-yy is the test Pi, and may be
					# inside the CAIDA/24 too. Do not drop those.

					if node != 'san-us' and node != 'xxx-yy':
						if log != None:
							log.flush()
						continue

				if have_state:

					if not state_db.node_has_state( node ):
						load_history ( state_db, node )

					if 'HWmac' in pkt and not state_db.node_has_state( node ):

						# If the first packet send from a node after a hostname change contains
						# the mac address of the network interface, and the mac address already
						# is associated with a node in the state db, then that node is renamed.

						# Note that there is a problem here: if the existing node with the
						# mac address was inactive for a long time, it may have been removed
						# from the state db, and it will not be found here (the mac address
						# does still reside in the history db). In that case

						# node=pkt['node']: new hostname
						# pkt['Hostname' ]: old hostname

						tt, vv = state_db.find_node_for_value( 'HWmac', pkt['HWmac'] )
						if tt != None:
							say.say( 'new node %s has same mac as node %s'%(node, vv) )
							pkt['Hostname'] = vv	# Old hostname with same MAC

			if node != pkt['Hostname'] and have_state:
				message = "hostname change: '%s' -> '%s'"%(pkt['Hostname'], node)

				# Do not proceed with the hostname change from pkt['Hostname']
				# to node = pkt['node'] if node already has been used, i.e.
				# if it is in the state db already, or has a history file.
				# These data would be overwritten with pkt['Hostname'] data.

				if state_db.node_has_state( node ):
					say.say ( 'abort %s (%s in state db)'%(message, node) )
					continue

				if state_db.node_path:
					node_file = os.path.join( state_db.node_path, node+'.db' )
					if os.path.exists( node_file ):
						say.say ( 'abort %s (%s has history db)'%(message, node) )
						continue

				say.say( message )

				# pkt['Hostname'] is in the state db. This changes the hostname field
				# for all db entries to the new hostname node = pkt['node']

				state_db.change_hostname( pkt['Hostname'], posix_time, node )

			# Even if have_state = True, state_db.node_path might not be defined:
			# there could be a state db, but no directory with history dbs.

			if state_db.node_path:
				node_file = os.path.join( state_db.node_path, node+'.db' )

				if node != pkt['Hostname']:
					# pkt['Hostname'] is in the state db, so it must have a history file,
					# so always node_file_exists = True.

					old_node_file = os.path.join( state_db.node_path, pkt['Hostname']+'.db' )
					if os.path.exists( old_node_file ):		# Always true
						say.say( '%s -> %s'%(old_node_file, node_file))
						os.rename( old_node_file, node_file )

				node_file_exists = os.path.exists( node_file )
				node_db = HistoryDB( node_file, state_db.tbl_names, create_tables=True )
				if not node_file_exists:	# New history file created
					os.chmod( node_file, stat.S_IREAD | stat.S_IWRITE | stat.S_IRGRP | stat.S_IWGRP | stat.S_IROTH )
			else:
				node_db = None

			# Update db only if attributes change (incl. new activations)

			mon_count += 1
			current_nodes.add(node)

			changes = {}

			for tbl in state_db.tbl_names:

				if tbl not in pkt:							# pkt may not have key
					continue

				if state_db.tbl_layout[tbl]['live'] == 0:   # Skip if 'live' flag is off
					continue

				tt, vv = state_db.find_value_for_node( tbl, node ) if have_state else (None,None)
				new_data = False
				if pkt[tbl] == 'None':						# key value not defined
					if tt != None:							# If time of last change is defined ..
						say.say( '%-15s %7s (delete)'%(tbl, node) )
						if have_state:
							state_db.delete( tbl, node )	# then delete the record
						new_data = True
				else:										# key value defined

					# Write to state db and to history db if:
					# - key is PktType
					# - it is a boot packet: pkt['Boot']
					# - if the value changes

					if (tbl == 'PktType'				 ) or \
					   (tbl == 'Boot' and pkt[tbl] == '1') or \
					   (tbl != 'Boot' and (tt == None or pkt[tbl] != vv) ):	# Not in db yet, or value changed

						if tbl != 'Uptime' and tbl != 'PktType':			# Skip these: too many
							say.say( '%-15s %7s %s %s -> %s'%(tbl, node, posix_time, vv if vv else 'None', pkt[tbl]) )

						if tbl != 'Utils' or len(pkt[tbl]) == 16:			# Only accept Utils if it has 16 entries
							if have_state:
								state_db.insert( tbl, node, posix_time, pkt[tbl] )
							new_data = True

				if new_data:
					if tbl == 'PktType':
						vv = -1 if vv == None else int(vv)
					changes[tbl] = { 'old': vv, 'new': pkt[tbl] }
					if state_db.tbl_layout[tbl]['hist'] and node_db:
						node_db.insert( tbl, posix_time, pkt[tbl] )
				#elif state_db.tbl_layout[tbl]['hist'] and node_db.count(tbl) == 0:
				#	node_db.insert( tbl, posix_time, pkt[tbl] )

			if node_db:
				node_db.db.close()
				node_db = None

			if state_db.task_path:
				handle_changes( state_db.task_path, pkt, changes )

			if not have_state or (trigger_delta and posix_time-trigger_time > trigger_delta):

				trigger_time = posix_time

				births = current_nodes.difference(previous_nodes)
				deaths = previous_nodes.difference(current_nodes)

				if log:
					log.write('### %s (new: %s; gone: %s)\n'%(mon_count, ','.join(births), ','.join(deaths)))
					log.flush()

				if state_db.task_path:
					file = os.path.join( state_db.task_path, str(int(round(time())))+'.global.task' )
					open(file,'w').write( dumps( { 'nodes': list(current_nodes), 'births': list(births), 'deaths': list(deaths) } )+'\n' )

				previous_nodes = current_nodes
				current_nodes  = set()
				mon_count = 0

				# Clean up state db? Remove nodes that haven't been heard of for a long time

				if have_state:
					nodes = state_db.nodes()
					now = time()
					for node in nodes:
						tt, vv = state_db.find_value_for_node( 'Uptime', node )
						if tt == None or now-tt < 182*86400:
							continue

						say.say( "remove dead node '%s' from state DB (last ping more than 6 months ago)"%node )

						for tbl in state_db.tbl_names:
							tt, vv = state_db.find_value_for_node( tbl, node )
							if tt != None:
								state_db.delete( tbl, node )

					state_db.reload()

			pkt_count += 1
			if 0 < args.count <= pkt_count:
				break

	except:										# Usually after ctrl-C
		say.warn( 'packet capture aborted' )
		raise

	say.done()
