import os
import sqlite3
#from time import sleep
import libtiny
from ark_ping_layout import table_layout

class StateDB:

	def __init__( self, db_file, create_tables=False, force=False ):
		say = libtiny.say('StateDB::__init__')

		tbl_layout, tbl_names, tbl_task = table_layout()

		if not db_file or (say.DRYRUN and not force):
			say.debug( db_file if db_file else 'no db' )

			self.db			= None
			self.cursor		= None
			self.db_file	= db_file if db_file else None

		else:
			say.debug( db_file )

			# Fudge to deal with locking issues with sqlite3.
			# Try opening a couple of times before aborting.

			#retries = 0
			#while True:
			#	try:
			#		self.db = sqlite3.connect( db_file )
			#	except:
			#		retries += 1
			#		if retries <= 3:
			#			say.say( "%s: retry ..."%db_file )
			#			sleep(1)
			#			continue
			#		raise
			#	else:
			#		break

			self.db = sqlite3.connect( db_file )
			self.cursor = self.db.cursor()
			self.db_file = db_file

		if self.db_file:
			path = os.path.join( os.path.dirname( self.db_file ), 'node-db' )
			self.node_path = path if os.path.isdir( path ) else None

			path = os.path.join( os.path.dirname( self.db_file ), 'task' )
			self.task_path = path if os.path.isdir( path ) else None
		else:
			self.node_path = None
			self.task_path = None

		self.reload( create_tables )

	def reload( self, create_tables=False ):
		say = libtiny.say('StateDB::reload')

		# Reloads the table layout and creates missing tables.

		self.tbl_layout, self.tbl_names, self.tbl_task = table_layout()

		if not say.DRYRUN and create_tables and self.db:
			for tbl in self.tbl_names:
				self.cursor.execute('''CREATE TABLE IF NOT EXISTS %s ( node TEXT NOT NULL PRIMARY KEY, time INTEGER NOT NULL, value TEXT NOT NULL )'''%tbl)
			self.db.commit()

	def change_hostname( self, node, time, new_node ):
		say = libtiny.say('StateDB::change_hostname')

		if say.DRYRUN:
			say.yell ('%s %s %s -> %s'%(node, time, node, new_node))
		else:
			say.debug('%s %s %s -> %s'%(node, time, node, new_node))
			for tbl in self.tbl_names:
				self.cursor.execute("DELETE FROM %s WHERE node='%s'"%(tbl, new_node))
				self.cursor.execute("UPDATE %s SET node='%s', time=%s WHERE node='%s'"%(tbl,new_node,time,node))
			self.db.commit()

	def insert( self, attr, node, time, value ):
		say = libtiny.say('StateDB::insert')

		if say.DRYRUN:
			say.yell('%s %s %s=%s'%(node, time, attr, value))
		else:
			say.debug('%s %s %s=%s'%(node, time, attr, value))
			try:
				self.cursor.execute('''INSERT OR REPLACE INTO %s(node, time, value) VALUES(?, ?, ?)'''%attr, (node, time, '%s'%value))
			except:
				say.yell( "node: %s, time: %s, attr: %s, value: %s"%(node, time, attr, value) )
				raise
			self.db.commit()

	def delete( self, attr, node ):
		say = libtiny.say('StateDB::delete')

		message = '%s %s'%(node, attr)
		if say.DRYRUN:
			say.yell( message )
		else:
			say.debug( message )
			self.cursor.execute("DELETE FROM %s WHERE node='%s'"%(attr, node))
			self.db.commit()

	# find_value_for_node (attr, node)							use state db only
	# find_value_for_node (attr, node, node_db)					use state db, then history
	# find_value_for_node (attr, node, node_db, position=0)		use state db only
	# find_value_for_node (attr, node, node_db, position=1)		use state db, then history
	#
	# use HistoryDB:: find_value_for_node to look through history only

	def find_value_for_node( self, attr, node, node_db=None, position=1 ):
		say = libtiny.say('StateDB::find_value_for_node')

		try:
			s = (node,)
			r = self.cursor.execute("SELECT time,value FROM %s where node=?"%attr, s).fetchone()
		except:
			r = None

		if r == None and node_db and position == 1:			# Try history
			r = node_db.find_value_for_node( attr, node, position=position )

		return [None, None] if r == None else r			# Time, value

	def find_node_for_value( self, attr, value, node_db=None ):
		say = libtiny.say('StateDB::find_node_for_value')

		try:
			s = (value,)
			r = self.cursor.execute("SELECT time,node FROM %s where value=?"%attr, s).fetchone()
		except:
			r = None

		# If nothing was found, try the history db

		if r == None and node_db:						# Pull most recent value from history
			s = node_db.dump( attr )

			s.sort( key=lambda m: m[0], reverse=True )	# Reverse chronological
			for x in s:
				if x[1] == value:
					r = x
					say.debug( 'from history: %s=%s'%r )
					break

		return [None, None] if r == None else r			# Time, node

	def values( self, attr ):
		say = libtiny.say('StateDB::values')

		try:		# Triggered when table does not exist
			rows = sorted( self.cursor.execute("SELECT value FROM %s"%attr), key=lambda name: name[0])
		except:
			rows = []

		return rows

	def nodes( self, all_nodes=False ):
		say = libtiny.say('StateDB::nodes')

		# Nodes in status db

		names = [ row[0] for row in self.cursor.execute("SELECT node FROM IPv4Global") ]

		# Add nodes in node_db. This adds nodes not in status db, because they
		# have not pinged home for a long time
		# (hence also includes decommissioned nodes)

		if all_nodes:
			db_names = []
			db_file  = self.db_file
			for f in os.listdir( os.path.join( os.path.dirname(db_file), 'node-db' ) ):
				if f != 'decommissioned':
					db_names.append( os.path.splitext( f )[0] )
			names = list( set(db_names) | set(names) )

		names = sorted( names )

		return names

	def node_has_state (db, node):
		say = libtiny.say('StateDB::node_has_state')

		# Every node in the state db should have an IPv4Global address
		# (or an IPv6Global address; will not happen until the nodes
		# start pinging home over ipv6).

		# There should not be any other values in the db at all if
		# there is no IPv4Global, but we are not checking for this.

		tt, vv = db.find_value_for_node( 'IPv4Global', node )
		if tt == None:
			return False
			#tt, vv = db.find_value_for_node( 'IPv6Global', node )
			#if tt == None:
			#	return False

		return True

	def dump( self, attr, posix=False ):
		say = libtiny.say('StateDB::dump')

		print ("# %s"%attr)
		if posix:
			for row in sorted( self.cursor.execute("SELECT node,time,value FROM %s"%attr), key=lambda name: name[0]):
				print ('%-7s %-40s %s'%(row[0],row[2],row[1]))
		else:
			for row in sorted( self.cursor.execute("SELECT node,time,value FROM %s"%attr), key=lambda name: name[0]):
				print ('%-7s %-40s %s'%(row[0],row[2],eon_date(posix=row[1]).get('YYYY-MM-DDThh:mm:ss')))

	def table_exists(self, attr ):
		say = libtiny.say('StateDB::table_exists')

		r = self.cursor.execute("SELECT count(*) FROM sqlite_master WHERE type='table' AND name='%s';"%attr).fetchone()
		return True if r[0] == 1 else False

	def delete_table( self, attr ):
		say = libtiny.say('StateDB::delete_table')

		if say.DRYRUN:
			say.yell("delete table '%s' from '%s'"%(attr,self.db_file))
		elif self.table_exists( attr ):
			say.say("delete table '%s' from '%s'"%(attr,self.db_file))
			self.cursor.execute("DROP TABLE IF EXISTS %s"%attr)
			self.db.commit()
		else:
			say.say("table '%s' does not exist in '%s'"%(attr,self.db_file))


class HistoryDB:

	def __init__( self, db_file, tbl_names=[], create_tables=False, force=False ):
		say = libtiny.say('HistoryDB::__init__')
		say.debug( db_file )

		if not db_file or (say.DRYRUN and not force):
			self.db		 = None
			self.cursor  = None
			self.db_file = db_file if db_file else None

		else:
			self.db		 = sqlite3.connect( db_file )
			self.cursor  = self.db.cursor()
			self.db_file = db_file

			if not say.DRYRUN and create_tables:
				for tbl in tbl_names:
					self.cursor.execute('''CREATE TABLE IF NOT EXISTS %s ( time INTEGER NOT NULL PRIMARY KEY, value TEXT NOT NULL )'''%tbl)
				self.db.commit()

		if self.db_file:
			self.node = os.path.splitext(os.path.basename(self.db_file))[0]

	def insert( self, attr, time, value ):
		say = libtiny.say('HistoryDB::insert into %s for %s'%(attr, self.node))
		message = '%s %s=%s'%(time, attr, value)

		if say.DRYRUN:
			say.yell( message )
		else:
			say.debug( message )
			try:
				self.cursor.execute('''INSERT OR REPLACE INTO %s(time, value) VALUES(?, ?)'''%attr, (time, '%s'%value))
			except:
				say.yell( "node: %s, %s"%(self.node, message) )
				raise
			self.db.commit()

	def find_value_for_node( self, attr, node, position=1 ):
		say = libtiny.say('HistoryDB::find_value_for_node for %s in %s'%(attr, self.node))

		r = None
		s = self.dump( attr )	# Returned in chronological order
		if len(s) >= position:
			#s.sort( key=lambda m: m[0] )
			r = s[-position]

		return [None, None] if r == None else r			# Time, value

	def delete_time( self, attr, time ):
		say = libtiny.say('HistoryDB::delete_time from %s in %s'%(attr, self.node))

		rows = [ row for row in self.cursor.execute("SELECT time,value FROM %s WHERE time=%s"%(attr, time)) ]
		message = '%s %s=%s'%(time, attr, ','.join( [row[1] for row in rows] ))

		if rows:
			if say.DRYRUN:
				say.yell( message )
			else:
				say.debug( message )
				try:
					self.cursor.execute("DELETE FROM %s WHERE time=%s"%(attr, time))
				except:
					say.yell( "node: %s, %s"%(self.node, message) )
					raise
				self.db.commit()
		else:
			say.debug( message )

	def count( self, attr ):
		say = libtiny.say('HistoryDB::count %s in %s'%(attr, self.node))

		try:
			rows = [ t[0] for t in self.cursor.execute("SELECT time FROM %s"%attr) ]
		except:
			rows = []

		return len( rows )

	def dump( self, attr, start_time=None, stop_time=None ):
		say = libtiny.say('HistoryDB::delete_time from %s in %s'%(attr, self.node))

		# Chronologically sorted (time,value) pairs
		try:
			rows = sorted( self.cursor.execute("SELECT time,value FROM %s"%attr), key=lambda row: row[0])
		except:
			rows = []

		rows = [ row for row in rows if (start_time == None or start_time <= row[0]) and (stop_time == None or row[0] < stop_time) ]

		return rows

	def table_exists(self, attr ):
		say = libtiny.say('HistoryDB::table_exists for %s in %s'%(attr, self.node))

		r = self.cursor.execute("SELECT count(*) FROM sqlite_master WHERE type='table' AND name='%s';"%attr).fetchone()
		return True if r[0] == 1 else False

	def delete_table( self, attr ):
		say = libtiny.say('HistoryDB::delete from %s in $s'%(attr, self_node))

		if say.DRYRUN:
			say.yell("delete table '%s' from '%s'"%(attr,self.db_file))
		elif self.table_exists(attr ):
			say.say("delete table '%s' from '%s'"%(attr,self.db_file))
			self.cursor.execute("DROP TABLE IF EXISTS %s"%attr)
			self.db.commit()
		else:
			say.say("table '%s' does not exist in '%s'"%(attr,self.db_file))
